# Tai Sakuma <tai.sakuma@gmail.com>
import logging
import pytest

from qtwirl._parser.expander import _expand_one_dict

##__________________________________________________________________||
params = [
    pytest.param(
        dict(abc_cfg=dict(A=1)),
        dict(
            config_keys=['abc_cfg'],
            default_config_key='abc_cfg',
        ),
        dict(abc_cfg=dict(expanded=dict(A=1))),
        id='simple'
    ),

    pytest.param(
        dict(A=1), # without config key, e.g, 'abc_cfg'
        dict(
            config_keys=['abc_cfg'],
            default_config_key='abc_cfg',
        ),
        dict(abc_cfg=dict(expanded=dict(A=1))),
        id='default'
    ),

    pytest.param(
        dict(xyz_cfg=dict(A=1)), # 'xyz_cfg' is not a config key
        dict(
            config_keys=['abc_cfg'],
            default_config_key='abc_cfg',
        ),
        dict(abc_cfg=dict(expanded=dict(xyz_cfg=dict(A=1)))),
        id='default'
    ),

    pytest.param(
        dict(xyz_cfg=dict(A=1)), # 'xyz_cfg' is a config key
        dict(
            config_keys=['abc_cfg', 'xyz_cfg'],
            default_config_key='abc_cfg',
        ),
        dict(xyz_cfg=dict(A=1)),
        id='no-expansion'
    ),

    pytest.param(
        dict(xyz_cfg=dict(A=1)),
        dict(
            config_keys=['abc_cfg'],
            default_config_key=None, # default is None
        ),
        dict(xyz_cfg=dict(A=1)),
        id='no-default'
    ),

    pytest.param(
        dict(A=1),
        dict(
            config_keys=['abc_cfg', 'xyz_cfg'],
            default_config_key='xyz_cfg',
        ),
        dict(xyz_cfg=dict(A=1)), # not expanded
        id='default-no-expansion'
    ),

    pytest.param(
        dict(A=1),
        dict(
            config_keys=['abc_cfg', 'def_cfg'],
            default_config_key='def_cfg',
        ),
        dict(def_cfg=dict(expanded=dict(A=1))),
        id='no-shared-option' # expand_def_cfg() doesn't take `shared`
    ),

]

def expand_abc_cfg(cfg, shared):
    return dict(abc_cfg=dict(expanded=cfg))

def expand_def_cfg(cfg): # doesn't take shared
    return dict(def_cfg=dict(expanded=cfg))

@pytest.mark.parametrize('cfg, shared, expected', params)
def test_expand_one_dict(cfg, shared, expected):
    shared['expand_func_map'] = {
        'abc_cfg': expand_abc_cfg,
        'def_cfg': expand_def_cfg,
        }
    actual = _expand_one_dict(cfg, shared)
    assert expected == actual
    assert actual is not cfg

##__________________________________________________________________||
params = [
    pytest.param(
        dict(abc_cfg=dict(A=1)),
        dict(
            config_keys=['abc_cfg'],
            default_config_key='abc_cfg',
            expand_func_map={'abc_cfg': expand_abc_cfg},
        ),
        dict(abc_cfg=dict(expanded=dict(A=1))), # not applied, only expanded
        id='with-expand-func'
    ),

    pytest.param(
        dict(abc_cfg=dict(A=1)),
        dict(
            config_keys=['abc_cfg'],
            default_config_key='abc_cfg',
            expand_func_map={ },
        ),
        dict(abc_cfg=dict(shared_applied=dict(A=1))),  # not expanded, only applied
        id='without-expand-func'
    ),

    pytest.param(
        dict(abc_cfg=dict(A=1)),
        dict(
            config_keys=[ ],
            default_config_key=None,
            expand_func_map={ },
        ),
        dict(abc_cfg=dict(shared_applied=dict(A=1))), # not applied, only expanded
        id='no-default'
    ),

    pytest.param(
        dict(abc_cfg=dict(A=1), xyz_cfg=dict(A=1)),
        dict(
            config_keys=[ ],
            default_config_key=None,
            expand_func_map={ },
        ),
        dict(abc_cfg=dict(A=1), xyz_cfg=dict(A=1)), # not applied, or expanded
        id='two-items'
    ),

]

def mock_apply_default_for_all_keys(cfg, shared):
    ret = cfg.copy()
    if 'abc_cfg' in ret:
        ret['abc_cfg'] = dict(shared_applied=ret['abc_cfg'])
    return ret

@pytest.fixture()
def monkeypatch_apply_default_for_all_keys(monkeypatch):
    import qtwirl._parser.expander as module
    monkeypatch.setattr(module, '_apply_default_for_all_keys', mock_apply_default_for_all_keys)
    return

@pytest.mark.parametrize('cfg, shared, expected', params)
def test_expand_one_dict_apply(cfg, shared, expected, monkeypatch_apply_default_for_all_keys):
    actual = _expand_one_dict(cfg, shared)
    assert expected == actual
    assert actual is not cfg

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import logging
import pytest

from qtwirl._parser.expander import _apply_default_for_all_keys

##__________________________________________________________________||
params = [

    pytest.param(
        dict(abc_cfg=dict(A=1), def_cfg=dict(D=2), set_default=dict(H=22)),
        dict(
            default_cfg_stack=[
                dict(abc_cfg=dict(A=5, B=3, C=10), def_cfg=dict(D=5, E=3)),
                dict(abc_cfg=dict(B=2)),
                dict(set_default=dict(G=20)),
            ],
        ),
        dict(abc_cfg=dict(A=1, B=2, C=10), def_cfg=dict(D=2, E=3), set_default=dict(H=22)),
        id='simple'
    ),

    pytest.param(
        dict(abc_cfg=dict(A=1), def_cfg=dict(D=2), set_default=dict(H=22)),
        dict(),
        dict(abc_cfg=dict(A=1), def_cfg=dict(D=2), set_default=dict(H=22)),
        id='no-default-cfg-stack'
    ),

]

param_names = 'cfg, shared, expected'

@pytest.mark.parametrize(param_names, params)
def test_apply_default_for_all_keys(cfg, shared, expected):
    assert expected == _apply_default_for_all_keys(cfg, shared)

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import pytest

try:
    import unittest.mock as mock
except ImportError:
    import mock

from qtwirl._parser.expander import _expand_config

##__________________________________________________________________||
def test_expand_config_none():
    cfg = None
    expected = None
    actual = _expand_config(cfg)
    assert expected == actual

##__________________________________________________________________||
@pytest.fixture()
def mock_expand_one_dict(monkeypatch):
    ret = mock.Mock()
    import qtwirl._parser.expander as module
    monkeypatch.setattr(module, '_expand_one_dict', ret)
    return ret

@pytest.mark.parametrize('cfg', [dict(), dict(A=1)])
@pytest.mark.parametrize('func_ret', [None, [], {}, dict(A=1), [dict(B=2, C=3)]])
def test_expand_config_one_dict(cfg, func_ret, mock_expand_one_dict):
    if func_ret:
        expected = func_ret
    else:
        expected = None
    mock_expand_one_dict.return_value = func_ret
    actual = _expand_config(cfg)
    assert expected == actual

##__________________________________________________________________||
@pytest.mark.parametrize(
    'cfg, func_side_eff, expected', [
        ([], [], []),
        ([None],[None], []),
        ([None],[dict(X=1)], []),
        ([None],[[]], []),
        ([None],[[dict(X=1)]], []),
        ([None],[[dict(X=1), dict(Y=2)]], []),
        ([dict(A=1)], [None], []),
        ([dict(A=1)], [dict(X=1)], [dict(X=1)]),
        ([dict(A=1)], [[]], []),
        ([dict(A=1)], [[dict(X=1)]], [dict(X=1)]),
        ([dict(A=1)], [[dict(X=1), dict(Y=2)]], [dict(X=1), dict(Y=2)]),
        ([dict(A=1)], [[dict(X=1), dict(Y=2), None]], [dict(X=1), dict(Y=2)]),

        ([dict(A=1), None], [None, dict(X=1)], []),
        ([dict(A=1), None], [dict(X=1), dict(X=1)], [dict(X=1)]),
        ([dict(A=1), None], [[], dict(X=1)], []),
        ([dict(A=1), None], [[dict(X=1)], dict(X=1)], [dict(X=1)]),
        ([dict(A=1), None], [[dict(X=1), dict(Y=2)], dict(X=1)], [dict(X=1), dict(Y=2)]),
        ([dict(A=1), None], [[dict(X=1), dict(Y=2), None], dict(X=1)], [dict(X=1), dict(Y=2)]),

        ([dict(A=1), dict(B=2)], [None, None], []),
        ([dict(A=1), dict(B=2)], [dict(X=1), None], [dict(X=1)]),
        ([dict(A=1), dict(B=2)], [dict(X=1), dict(Y=2)], [dict(X=1), dict(Y=2)]),
        ([dict(A=1), dict(B=2)], [dict(X=1), [dict(Y=2), dict(Z=3)]], [dict(X=1), dict(Y=2), dict(Z=3)]),
        ([dict(A=1), dict(B=2)], [dict(X=1), [dict(Y=2), dict(Z=3), None]], [dict(X=1), dict(Y=2), dict(Z=3)]),

    ]
)
def test_expand_config_list(cfg, func_side_eff, expected, mock_expand_one_dict):
    mock_expand_one_dict.side_effect = func_side_eff
    actual = _expand_config(cfg)
    assert expected == actual

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import pytest

from qtwirl._parser.expander import config_expander

##__________________________________________________________________||
def expand_abc_cfg(cfg, shared):
    key = 'abc_cfg'
    cfg = shared['func_apply_default'](key, cfg)
    return {key: dict(expanded=cfg)}

##__________________________________________________________________||
def test_config_expander():
    expand_func_map = {
        'abc_cfg': expand_abc_cfg,
    }
    config_keys = ['def_cfg']
    default_config_key = 'abc_cfg'
    expand_config=config_expander(
        expand_func_map=expand_func_map,
        config_keys=config_keys,
        default_config_key=default_config_key
    )

    cfg = [
        dict(set_default=dict(
            abc_cfg=dict(A=5, C=4),
            def_cfg=dict(D=2, E=9),
        )),
        dict(A=1, B=2)
    ]

    expected = [{'abc_cfg': {'expanded': dict(A=1, B=2, C=4)}}]
    assert expected == expand_config(cfg)

    cfg = [
        dict(set_default=dict(A=5, C=4)),
        dict(A=1, B=2)
    ]

    expected = [{'abc_cfg': {'expanded': dict(A=1, B=2, C=4)}}]
    assert expected == expand_config(cfg)

    cfg = [
        dict(set_default=dict(abc_cfg=dict(B=3, C=2, E=8))),
        dict(set_default=dict(A=5, C=4)),
        dict(A=1, B=2),
    ]
    expected = [
        {'abc_cfg': {'expanded': dict(A=1, B=2, C=4, E=8)}},
    ]
    assert expected == expand_config(cfg)

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import pytest

from qtwirl._parser.tableconfig import compose_tbl_filename_from_config

##__________________________________________________________________||
params = [
    pytest.param(
        [dict(key_out_name=('var1', 'var2'))],
        { }, 'tbl.var1.var2.txt', id='simple'),
    pytest.param(
        [dict(key_out_name=('var1', 'var2'), file_name_prefix='def')],
        { }, 'def.var1.var2.txt', id='simple'),
    pytest.param(
        [dict(key_out_name=('var1', 'var2'))],
        dict(default=dict(file_name_prefix='abc')),
        'abc.var1.var2.txt', id='default'),
    pytest.param(
        [dict(key_out_name=('var1', 'var2'), file_name_prefix='def')],
        dict(default=dict(file_name_prefix='abc')),
        'def.var1.var2.txt', id='override'),

    pytest.param(
        [dict(key_out_name=('var1', 'var2'), key_index=(1, 2))],
        dict(
            default=dict(
                file_name_prefix='abc',
                file_name_suffix='csv',
                file_name_var_separator='#',
                file_name_idx_separator='=',
            )),
        'abc#var1=1#var2=2.csv', id='default-all'),

    pytest.param(
        [dict(
            key_out_name=('var1', 'var2'),
            key_index=(1, 2),
            file_name_prefix='def',
            file_name_suffix='dat',
            file_name_var_separator='+',
            file_name_idx_separator='&',
        )],
        dict(
            default=dict(
                file_name_prefix='abc',
                file_name_suffix='csv',
                file_name_var_separator='#',
                file_name_idx_separator='=',
            )),
        'def+var1&1+var2&2.dat', id='default-all-override'),
]

@pytest.mark.parametrize('args, kwargs, expected', params)
def test_complete(args, kwargs, expected):
    actual = compose_tbl_filename_from_config(*args, **kwargs)
    assert expected == actual

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import logging
import pytest

try:
    import unittest.mock as mock
except ImportError:
    import mock

from qtwirl._parser.expander import config_expander

##__________________________________________________________________||
def expand_abc_cfg(cfg):
    return dict(abc_cfg=dict(expanded=cfg))

def expand_def_cfg(cfg):
    return dict(def_cfg=dict(expanded=cfg))

##__________________________________________________________________||
def test_simple():
    expand_func_map = {'abc_cfg': expand_abc_cfg}
    config_keys = [ ]
    default_config_key = 'abc_cfg'
    expand_config = config_expander(
        expand_func_map=expand_func_map,
        config_keys=config_keys,
        default_config_key=default_config_key
    )

    assert {'abc_cfg': {'expanded': {'A': 1}}} == expand_config(dict(A=1))

##__________________________________________________________________||
def test_set_default_not_overridden(caplog):
    mock_set_default = mock.Mock()
    expand_func_map = {
        'abc_cfg': expand_abc_cfg,
        'set_default' : mock_set_default, # give 'set_default'
    }
    config_keys = [ ]
    default_config_key = 'abc_cfg'

    #
    with caplog.at_level(logging.INFO):
        expand_config = config_expander(
            expand_func_map=expand_func_map,
            config_keys=config_keys,
            default_config_key=default_config_key)

    assert len(caplog.records) == 1
    assert caplog.records[0].levelname == 'INFO'
    assert 'expander' in caplog.records[0].name
    assert '"set_default" is in' in caplog.records[0].msg

    shared = expand_config.keywords['shared']
    assert mock_set_default == shared['expand_func_map']['set_default']

##__________________________________________________________________||
@pytest.mark.parametrize('default_cfg_dict', [None, dict(A=1)])
def test_default_cfg_dict(default_cfg_dict):
    expand_config = config_expander(default_cfg_dict=default_cfg_dict)
    shared = expand_config.keywords['shared']
    if default_cfg_dict is None:
        assert [{}] == shared['default_cfg_stack']
    else:
        assert [default_cfg_dict] == shared['default_cfg_stack']

##__________________________________________________________________||
def test_func_apply_default():
    expand_config = config_expander()
    shared = expand_config.keywords['shared']
    func_apply_default = shared['func_apply_default']
    assert 'apply_default' == func_apply_default.__name__

##__________________________________________________________________||
def test_wrap():
    mock_set_default = mock.Mock()
    expand_func_map = {
        'abc_cfg': expand_abc_cfg,
        'def_cfg': expand_def_cfg,
    }
    config_keys = ['xyz_cfg']
    default_config_key = 'abc_cfg'

    #
    expand_config = config_expander(
        expand_func_map=expand_func_map,
        config_keys=config_keys,
        default_config_key=default_config_key)

    assert 'expand_config' == expand_config.__name__
    assert  'expand a config' in expand_config.__doc__

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import pytest

from qtwirl._parser.inputs import parse_data

##__________________________________________________________________||
@pytest.mark.parametrize('arg, expected', [
    pytest.param(['A.root', 'B.root'], ['A.root', 'B.root'], id='list'),
    pytest.param('A.root', ['A.root'], id='string'),
    pytest.param([ ], [ ], id='empty-list'),
    pytest.param('', [ ], id='empty-string'),
])
def test_parse_data(arg, expected):
    assert expected == parse_data(arg)

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import pytest

from qtwirl._parser.tableconfig import compose_tbl_filename

##__________________________________________________________________||
params = [
    pytest.param([( )], { }, 'tbl.txt', id='empty'),
    pytest.param(
        [('var1', 'var2', 'var3')], { },
        'tbl.var1.var2.var3.txt',
        id='no-indices'),
    pytest.param(
        [('var1', 'var2', 'var3')],
        dict(key_indices=(1, None, 2)),
        'tbl.var1-1.var2.var3-2.txt',
        id='simple'),
    pytest.param(
        [('var1', 'var2', 'var3')],
        dict(key_indices=(1, None, 2), prefix='tbl_Sum'),
        'tbl_Sum.var1-1.var2.var3-2.txt',
        id='prefix'),
    pytest.param(
        [('var1', 'var2', 'var3')],
        dict(key_indices=(1, None, 2), suffix='hdf5'),
        'tbl.var1-1.var2.var3-2.hdf5',
        id='suffix'),
    pytest.param(
        [('var1', 'var2', 'var3')],
        dict(key_indices=(1, None, '*')),
        'tbl.var1-1.var2.var3-w.txt',
        id='star'),
    pytest.param(
        [('var1', 'var2', 'var3', 'var4', 'var5')],
        dict(key_indices=(1, None, '*', '(*)', '\\1')),
        'tbl.var1-1.var2.var3-w.var4-wp.var5-b1.txt',
        id='backref'),
    pytest.param(
        [('var1', 'var2', 'var3', 'var4', 'var5')],
        dict(key_indices=(1, None, '*', '(*)', '\\1'), var_separator='#'),
        'tbl#var1-1#var2#var3-w#var4-wp#var5-b1.txt',
        id='var-separator'),
    pytest.param(
        [('var1', 'var2', 'var3', 'var4', 'var5')],
        dict(key_indices=(1, None, '*', '(*)', '\\1'), idx_separator='#'),
        'tbl.var1#1.var2.var3#w.var4#wp.var5#b1.txt',
        id='idx-separator'),
]

@pytest.mark.parametrize('args, kwargs, expected', params)
def test_complete(args, kwargs, expected):
    actual = compose_tbl_filename(*args, **kwargs)
    assert expected == actual

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import pytest

try:
    import unittest.mock as mock
except ImportError:
    import mock

import alphatwirl

from qtwirl._parser.readerconfig import expand_reader_config

##__________________________________________________________________||
RoundLog = mock.Mock()

tblcfg_dict1 = dict(
    key_name=('jet_pt', ),
    key_binning=(RoundLog(0.1, 100), ),
    key_index=('*', ),
)

tblcfg_dict2 = dict(
    key_name=('met', ),
    key_binning=(RoundLog(0.1, 100), ),
)

selection_cfg_dict = dict(All=('ev: ev.njets[0] > 4', ))
selection_cfg_str = 'ev: ev.njets[0] > 4'

scribbler1 = mock.MagicMock()
scribbler2 = mock.Mock()

##__________________________________________________________________||
params = [

    #
    pytest.param(None, None, id='none'),

    #
    ##
    pytest.param(dict(reader=[]), None, id='reader-empty'),
    pytest.param(dict(reader=[None]), None, id='reader-none'),
    pytest.param(dict(reader=[None, None]), None, id='reader-none'),

    ##
    pytest.param(
        dict(reader=scribbler1),
        dict(reader=scribbler1),
        id='reader-reader'),
    pytest.param(
        dict(reader=[scribbler1]),
        [dict(reader=scribbler1)],
        id='reader-reader'),
    pytest.param(
        dict(reader=[scribbler1, None]),
        [dict(reader=scribbler1)],
        id='reader-reader'),
    pytest.param(
        dict(reader=[scribbler1, scribbler2]),
        [dict(reader=scribbler1), dict(reader=scribbler2)],
        id='reader-reader'),
    pytest.param(
        dict(reader=[scribbler1, None, scribbler2]),
        [dict(reader=scribbler1), dict(reader=scribbler2)],
        id='reader-reader'),

    #
    ##
    pytest.param([ ], [ ], id='list-empty'),
    pytest.param([None], [ ], id='list-none'),
    pytest.param([None, None], [ ], id='list-none'),

    ##
    pytest.param([dict(reader=[])], [ ], id='list-reader'),
    pytest.param([dict(reader=[]), dict(reader=[])], [ ], id='list-reader'),

    pytest.param(
        [dict(reader=scribbler1)],
        [dict(reader=scribbler1)],
        id='list-reader'),
    pytest.param(
        [dict(reader=[scribbler1])],
        [dict(reader=scribbler1)],
        id='list-reader'),
    pytest.param(
        [dict(reader=[scribbler1, None])],
        [dict(reader=scribbler1)],
        id='list-reader'),
    pytest.param(
        [dict(reader=[scribbler1]), None],
        [dict(reader=scribbler1)],
        id='list-reader'),

    #
    pytest.param(
        [dict(reader=[scribbler1, scribbler2])],
        [dict(reader=scribbler1), dict(reader=scribbler2)],
        id='two-readers'
    ),
    pytest.param(
        [dict(reader=[scribbler1]), dict(reader=scribbler2)],
        [dict(reader=scribbler1), dict(reader=scribbler2)],
        id='two-readers'
    ),
]

@pytest.mark.parametrize('arg, expected', params)
def test_expand_reader_config(arg, expected):
    assert expected == expand_reader_config(arg)

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import logging
import pytest

from qtwirl._parser.expander import _set_default, _wrap_default_cfg

##__________________________________________________________________||
params = [
    pytest.param(
        ['abc_cfg', 'def_cfg'], # config_keys
        'abc_cfg', # default_config_key
        dict(abc_cfg=dict(A=1)), # cfg
        dict(abc_cfg=dict(A=1)), # expected
        id='simple'
    ),

    pytest.param(
        ['abc_cfg', 'def_cfg'], # config_keys
        'abc_cfg', # default_config_key
        dict(A=1), # cfg
        dict(abc_cfg=dict(A=1)), # expected
        id='wrapped-with-default'
    ),

    pytest.param(
        ['abc_cfg', 'def_cfg'], # config_keys
        'abc_cfg', # default_config_key
        dict(abc_cfg=dict(A=1), def_cfg=dict(D=1)), # cfg
        dict(abc_cfg=dict(A=1), def_cfg=dict(D=1)), # expected
        id='two-items-in-cfg'
    ),

    pytest.param(
        ['abc_cfg', 'def_cfg'], # config_keys
        'abc_cfg', # default_config_key
        dict(abc_cfg=dict(A=1), def_cfg=dict(D=1), xyz_cfg=dict(X=1)), # cfg
        dict(abc_cfg=dict(abc_cfg=dict(A=1), def_cfg=dict(D=1), xyz_cfg=dict(X=1))), # expected
        id='extra-item-wrapped-with-default'
    ),

    # default_config_key: None
    pytest.param(
        ['abc_cfg', 'def_cfg'], # config_keys
        None, # default_config_key
        dict(abc_cfg=dict(A=1)), # cfg
        dict(abc_cfg=dict(A=1)), # expected
        id='default-none'
    ),

    pytest.param(
        ['abc_cfg', 'def_cfg'], # config_keys
        None, # default_config_key
        dict(A=1), # cfg
        dict(A=1), # expected
        id='default-none'
    ),

    pytest.param(
        ['abc_cfg', 'def_cfg'], # config_keys
        None, # default_config_key
        dict(abc_cfg=dict(A=1), def_cfg=dict(D=1)), # cfg
        dict(abc_cfg=dict(A=1), def_cfg=dict(D=1)), # expected
        id='default-none'
    ),

    pytest.param(
        ['abc_cfg', 'def_cfg'], # config_keys
        None, # default_config_key
        dict(abc_cfg=dict(A=1), def_cfg=dict(D=1), xyz_cfg=dict(X=1)), # cfg
        dict(abc_cfg=dict(A=1), def_cfg=dict(D=1), xyz_cfg=dict(X=1)), # expected
        id='default-none'
    ),

]

param_names = 'config_keys, default_config_key, cfg, expected'

@pytest.mark.parametrize(param_names, params)
def test_set_default(config_keys, default_config_key, cfg, expected):
    default_cfg_stack = [ ]
    shared = dict(
        config_keys=config_keys,
        default_config_key=default_config_key,
        default_cfg_stack=default_cfg_stack
    )

    _set_default(cfg, shared)
    assert [expected] == default_cfg_stack

@pytest.mark.parametrize(param_names, params)
def test_wrap_default_cfg(config_keys, default_config_key, cfg, expected):
    assert expected == _wrap_default_cfg(cfg, config_keys, default_config_key)

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import pytest

try:
    import unittest.mock as mock
except ImportError:
    import mock

from qtwirl._parser.tableconfig import complete_table_cfg

from alphatwirl.summary import Count, WeightCalculatorOne


def eq(self, other):
    return isinstance(other, self.__class__)
WeightCalculatorOne.__eq__ = eq


##__________________________________________________________________||
class MockSummary2:
    pass

class MockBinning:
    pass

defaultWeight = WeightCalculatorOne()
binning1 = MockBinning()
binning2 = MockBinning()

##__________________________________________________________________||
params = [

    ## no key
    pytest.param(
        dict(),
        dict(
            key_name=(),
            key_index=None,
            key_binning=None,
            key_out_name=(),
            val_name=None,
            val_index=None,
            agg_class=Count,
            agg_name=('n', 'nvar'),
            weight=defaultWeight,
            sort=True,
            nevents=None,
            store_file=False,
        ),
        id='no-key-empty-dict'
    ),
    pytest.param(
        dict(key_name=( ), key_binning=( )),
        dict(
            key_name=(),
            key_index=None,
            key_binning=(),
            key_out_name=(),
            val_name=None,
            val_index=None,
            agg_class=Count,
            agg_name=('n', 'nvar'),
            weight=defaultWeight,
            sort=True,
            nevents=None,
            store_file=False,
        ),
        id='no-key-empty-key'
    ),

    ## one key
    pytest.param(
        dict(
            key_name='met_pt',
        ),
        dict(
            key_name=('met_pt',),
            key_index=None,
            key_binning=None,
            key_out_name=('met_pt',),
            val_name=None,
            val_index=None,
            agg_class=Count,
            agg_name=('n', 'nvar'),
            weight=defaultWeight,
            sort=True,
            nevents=None,
            store_file=False,
        ),
        id='one-key-no-tuple'
    ),
    pytest.param(
        dict(
            key_name='met_pt',
            key_binning=binning1,
        ),
        dict(
            key_name=('met_pt',),
            key_index=None,
            key_binning=(binning1, ),
            key_out_name=('met_pt',),
            val_name=None,
            val_index=None,
            agg_class=Count,
            agg_name=('n', 'nvar'),
            weight=defaultWeight,
            sort=True,
            nevents=None,
            store_file=False,
        ),
        id='one-key-one-bin-no-tuple'
    ),
    pytest.param(
        dict(
            key_name='met_pt',
            key_index=1,
            key_binning=binning1,
            key_out_name='met',
        ),
        dict(
            key_name=('met_pt',),
            key_index=(1, ),
            key_binning=(binning1, ),
            key_out_name=('met',),
            val_name=None,
            val_index=None,
            agg_class=Count,
            agg_name=('n', 'nvar'),
            weight=defaultWeight,
            sort=True,
            nevents=None,
            store_file=False,
        ),
        id='one-key-idx-bin-out-no-tuple'
    ),
    pytest.param(
        dict(
            key_name=('met_pt', ),
            key_binning=(binning1, ),
        ),
        dict(
            key_name=('met_pt',),
            key_index=None,
            key_binning=(binning1, ),
            key_out_name=('met_pt',),
            val_name=None,
            val_index=None,
            agg_class=Count,
            agg_name=('n', 'nvar'),
            weight=defaultWeight,
            sort=True,
            nevents=None,
            store_file=False,
        ),
        id='one-key-tuple'
    ),

    ## summary class
    pytest.param(
        dict(
            key_name=( ),
            key_binning=( ),
            agg_class=MockSummary2,
        ),
        dict(
            key_name=(),
            key_index=None,
            key_binning=(),
            key_out_name=(),
            val_name=None,
            val_index=None,
            agg_class=MockSummary2,
            agg_name=(),
            weight=defaultWeight,
            sort=True,
            nevents=None,
            store_file=False,
        ),
        id='summary-class-empty-key-empty-val'
    ),
    pytest.param(
        dict(
            key_name=('key1', 'key2'),
            key_binning=(binning1, binning2),
            agg_class=MockSummary2,
        ),
        dict(
            key_name=('key1', 'key2'),
            key_index=None,
            key_binning=(binning1, binning2),
            key_out_name=('key1', 'key2'),
            val_name=None,
            val_index=None,
            agg_class=MockSummary2,
            agg_name=(),
            weight=defaultWeight,
            sort=True,
            nevents=None,
            store_file=False,
        ),
        id='summary-class-keys-empty-vals'
    ),
    pytest.param(
        dict(
            key_name=('key1', 'key2'),
            key_binning=(binning1, binning2),
            val_name=('val1', ),
            agg_class=MockSummary2,
        ),
        dict(
            key_name=('key1', 'key2'),
            key_index=None,
            key_binning=(binning1, binning2),
            key_out_name=('key1', 'key2'),
            val_name=('val1', ),
            val_index=None,
            agg_class=MockSummary2,
            agg_name=('val1', ),
            weight=defaultWeight,
            sort=True,
            nevents=None,
            store_file=False,
        ),
        id='summary-class-keys-one-val-tuple'
    ),
    pytest.param(
        dict(
            key_name=('key1', 'key2'),
            key_binning=(binning1, binning2),
            val_name=('val1', ),
            val_index=('*', ),
            agg_class=MockSummary2,
        ),
        dict(
            key_name=('key1', 'key2'),
            key_index=None,
            key_binning=(binning1, binning2),
            key_out_name=('key1', 'key2'),
            val_name=('val1', ),
            val_index=('*', ),
            agg_class=MockSummary2,
            agg_name=('val1', ),
            weight=defaultWeight,
            sort=True,
            nevents=None,
            store_file=False,
        ),
        id='summary-class-keys-one-val-idx-tuple'
    ),
    pytest.param(
        dict(
            key_name=('key1', 'key2'),
            key_binning=(binning1, binning2),
            val_name=('val1', 'val2'),
            agg_class=MockSummary2,
            agg_name=('agg1', ),
        ),
        dict(
            key_name=('key1', 'key2'),
            key_index=None,
            key_binning=(binning1, binning2),
            key_out_name=('key1', 'key2'),
            val_name=('val1', 'val2'),
            val_index=None,
            agg_class=MockSummary2,
            agg_name=('agg1', ),
            weight=defaultWeight,
            sort=True,
            nevents=None,
            store_file=False,
        ),
        id='summary-class-one-agg-tuple'
    ),
    pytest.param(
        dict(
            key_name=('key1', 'key2'),
            key_binning=(binning1, binning2),
            val_name=('val1', 'val2'),
            agg_class=MockSummary2,
            agg_name='agg1',
        ),
        dict(
            key_name=('key1', 'key2'),
            key_index=None,
            key_binning=(binning1, binning2),
            key_out_name=('key1', 'key2'),
            val_name=('val1', 'val2'),
            val_index=None,
            agg_class=MockSummary2,
            agg_name=('agg1', ),
            weight=defaultWeight,
            sort=True,
            nevents=None,
            store_file=False,
        ),
        id='summary-class-one-agg-no-tuple'
    ),
    pytest.param(
        dict(
            key_name=('key1', 'key2'),
            key_binning=(binning1, binning2),
            val_name='val1',
            agg_class=MockSummary2,
        ),
        dict(
            key_name=('key1', 'key2'),
            key_index=None,
            key_binning=(binning1, binning2),
            key_out_name=('key1', 'key2'),
            val_name=('val1', ),
            val_index=None,
            agg_class=MockSummary2,
            agg_name=('val1', ),
            weight=defaultWeight,
            sort=True,
            nevents=None,
            store_file=False,
        ),
        id='summary-class-keys-one-val-no-tuple'
    ),
    pytest.param(
        dict(
            key_name=('key1', 'key2'),
            key_binning=(binning1, binning2),
            val_name='val1',
            val_index='*',
            agg_class=MockSummary2,
        ),
        dict(
            key_name=('key1', 'key2'),
            key_index=None,
            key_binning=(binning1, binning2),
            key_out_name=('key1', 'key2'),
            val_name=('val1', ),
            val_index=('*', ),
            agg_class=MockSummary2,
            agg_name=('val1', ),
            weight=defaultWeight,
            sort=True,
            nevents=None,
            store_file=False,
        ),
        id='summary-class-keys-one-val-idx-no-tuple'
    ),
    pytest.param(
        dict(
            key_name=('key1', 'key2'),
            key_binning=(binning1, binning2),
            val_name=('val1', 'val2'),
            agg_class=MockSummary2,
        ),
        dict(
            key_name=('key1', 'key2'),
            key_index=None,
            key_binning=(binning1, binning2),
            key_out_name=('key1', 'key2'),
            val_name=('val1', 'val2'),
            val_index=None,
            agg_class=MockSummary2,
            agg_name=('val1', 'val2'),
            weight=defaultWeight,
            sort=True,
            nevents=None,
            store_file=False,
        ),
        id='summary-class-keys-2-vals'
    ),
    pytest.param(
        dict(
            key_name=('key1', 'key2'),
            key_binning=(binning1, binning2),
            key_index=(None, 1),
            val_name=('val1', 'val2'),
            agg_class=MockSummary2,
        ),
        dict(
            key_name=('key1', 'key2'),
            key_index=(None, 1),
            key_binning=(binning1, binning2),
            key_out_name=('key1', 'key2'),
            val_name=('val1', 'val2'),
            val_index=None,
            agg_class=MockSummary2,
            agg_name=('val1', 'val2'),
            weight=defaultWeight,
            sort=True,
            nevents=None,
            store_file=False,
        ),
        id='summary-class-2-keys-2-vals-key-indices'
    ),
    pytest.param(
        dict(
            key_name=('key1', 'key2'),
            key_binning=(binning1, binning2),
            val_name=('val1', 'val2'),
            agg_class=MockSummary2,
            val_index=(2, None),
        ),
        dict(
            key_name=('key1', 'key2'),
            key_index=None,
            key_binning=(binning1, binning2),
            key_out_name=('key1', 'key2'),
            val_name=('val1', 'val2'),
            val_index=(2, None),
            agg_class=MockSummary2,
            agg_name=('val1', 'val2'),
            weight=defaultWeight,
            sort=True,
            nevents=None,
            store_file=False,
        ),
        id='summary-class-2-keys-2-vals-val-indices'
    ),
    pytest.param(
        dict(
            key_name=('key1', 'key2'),
            key_binning=(binning1, binning2),
            key_index=(None, 1),
            val_name=('val1', 'val2'),
            val_index=(2, 3),
            agg_class=MockSummary2,
        ),
        dict(
            key_name=('key1', 'key2'),
            key_index=(None, 1),
            key_binning=(binning1, binning2),
            key_out_name=('key1', 'key2'),
            val_name=('val1', 'val2'),
            val_index=(2, 3),
            agg_class=MockSummary2,
            agg_name=('val1', 'val2'),
            weight=defaultWeight,
            sort=True,
            nevents=None,
            store_file=False,
        ),
        id='summary-class-2-keys-2-vals-key-indices-val-indices'
    ),
]

@pytest.mark.parametrize('arg, expected', params)
def test_complete(arg, expected):
    actual = complete_table_cfg(arg)
    assert expected == actual
    assert arg is not actual

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import pytest

try:
    import unittest.mock as mock
except ImportError:
    import mock

import alphatwirl

from qtwirl._parser.readerconfig import expand_reader_config

##__________________________________________________________________||
def mock_complete_table_cfg(cfg):
    return dict(mock_complete_table_cfg=cfg)

@pytest.fixture(autouse=True)
def monkeypatch_complete_table_cfg(monkeypatch):
    from qtwirl._parser import readerconfig
    monkeypatch.setattr(readerconfig, 'complete_table_cfg', mock_complete_table_cfg)

def mock_complete_selection_cfg(cfg):
    return dict(mock_complete_selection_cfg=cfg)

@pytest.fixture(autouse=True)
def monkeypatch_complete_selection_cfg(monkeypatch):
    from qtwirl._parser import readerconfig
    monkeypatch.setattr(readerconfig, 'complete_selection_cfg', mock_complete_selection_cfg)

##__________________________________________________________________||
RoundLog = mock.Mock()

tblcfg_dict1 = dict(
    key_name='jet_pt',
    key_binning=RoundLog(0.1, 100),
    key_index='*',
)


tblcfg_dict2 = dict(
    key_name='met',
    key_binning=RoundLog(0.1, 100),
)

tblcfg_dict1_completed = dict(mock_complete_table_cfg=tblcfg_dict1)
tblcfg_dict2_completed = dict(mock_complete_table_cfg=tblcfg_dict2)

selection_cfg_dict = dict(All=('ev: ev.njets[0] > 4', ))
selection_cfg_str = 'ev: ev.njets[0] > 4'

selection_cfg_dict_completed = dict(mock_complete_selection_cfg=selection_cfg_dict)
selection_cfg_str_completed = dict(mock_complete_selection_cfg=selection_cfg_str)

scribbler1 = mock.Mock()

##__________________________________________________________________||
params = [
    pytest.param(
        dict(), dict(table_cfg=dict(mock_complete_table_cfg=dict())), id='empty-dict'
    ),
    pytest.param(
        [ ], [ ], id='empty-list'
    ),
    pytest.param(
        dict(tblcfg_dict1),
        dict(table_cfg=tblcfg_dict1_completed),
        id='dict-short'
    ),
    pytest.param(
        dict(table_cfg=tblcfg_dict1),
        dict(table_cfg=tblcfg_dict1_completed),
        id='dict-full'
    ),
    pytest.param(
        [tblcfg_dict1],
        [dict(table_cfg=tblcfg_dict1_completed)],
        id='list-with-one-dict'
    ),
    pytest.param(
        [tblcfg_dict1, tblcfg_dict2],
        [
            dict(table_cfg=tblcfg_dict1_completed),
            dict(table_cfg=tblcfg_dict2_completed),
        ],
        id='two-table-cfgs-short'
    ),
    pytest.param(
        [
            dict(table_cfg=tblcfg_dict1),
            dict(tblcfg_dict2),
        ],
        [
            dict(table_cfg=tblcfg_dict1_completed),
            dict(table_cfg=tblcfg_dict2_completed),
        ],
        id='two-table-cfgs-full-short'
    ),
    pytest.param(
        [
            dict(table_cfg=tblcfg_dict1),
            dict(table_cfg=tblcfg_dict2),
        ],
        [
            dict(table_cfg=tblcfg_dict1_completed),
            dict(table_cfg=tblcfg_dict2_completed),
        ],
        id='two-table-cfgs-full'
    ),
    pytest.param(
        [
            dict(selection_cfg=selection_cfg_dict),
            tblcfg_dict1,
        ],
        [
            dict(selection_cfg=selection_cfg_dict_completed),
            dict(table_cfg=tblcfg_dict1_completed),
        ],
        id='one-selection-dict-one-table'
    ),
    pytest.param(
        [
            dict(selection_cfg=selection_cfg_str),
            tblcfg_dict1,
        ],
        [
            dict(selection_cfg=selection_cfg_str_completed),
            dict(table_cfg=tblcfg_dict1_completed),
        ],
        id='one-selection-str-one-table'
    ),
    pytest.param(
        dict(selection_cfg=selection_cfg_dict),
        dict(selection_cfg=selection_cfg_dict_completed),
        id='one-selection-dict'
    ),
    pytest.param(
        dict(selection_cfg=selection_cfg_str),
        dict(selection_cfg=selection_cfg_str_completed),
        id='one-selection-str'
    ),
    pytest.param(
        [
            dict(reader=scribbler1),
            tblcfg_dict1,
        ],
        [
            dict(reader=scribbler1),
            dict(table_cfg=tblcfg_dict1_completed),
        ],
        id='one-scribbler-one-table'
    ),
]

@pytest.mark.parametrize('arg, expected', params)
def test_expand_reader_config(arg, expected):
    actual = expand_reader_config(arg)
    assert expected == actual

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>

##__________________________________________________________________||
def is_dict(obj):
    try:
        # check for mixin methods of mapping
        # https://docs.python.org/3.6/library/collections.abc.html
        obj.keys()
        obj.items()
        obj.values()
    except AttributeError:
        return False

    return True

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import pytest

from qtwirl._parser.selectionconfig import complete_selection_cfg

##__________________________________________________________________||
params = [

    #
    (
        'ev: ev.njets[0] > 4',
        dict(condition='ev: ev.njets[0] > 4',
             count=False, store_file=False)
    ),
    (
        dict(All=('ev: ev.njets[0] > 4', )),
        dict(condition=dict(All=('ev: ev.njets[0] > 4', )),
             count=False, store_file=False)
    ),

    #
    (
        dict(condition='ev: ev.njets[0] > 4'),
        dict(condition='ev: ev.njets[0] > 4',
             count=False, store_file=False)
    ),
    (
        dict(condition=dict(All=('ev: ev.njets[0] > 4', ))),
        dict(condition=dict(All=('ev: ev.njets[0] > 4', )),
             count=False, store_file=False)
    ),

    #
    (
        dict(condition=dict(All=('ev: ev.njets[0] > 4', )),
             count=True, store_file=True),
        dict(condition=dict(All=('ev: ev.njets[0] > 4', )),
             count=True, store_file=True,
             file_path='tbl_selection_count.txt')
    ),

    #
    (
        dict(condition=dict(All=('ev: ev.njets[0] > 4', )),
             count=True, store_file=True,
             file_dir='out'),
        dict(condition=dict(All=('ev: ev.njets[0] > 4', )),
             count=True, store_file=True,
             file_dir='out',
             file_path='out/tbl_selection_count.txt')
    ),

    #
    (
        dict(condition=dict(All=('ev: ev.njets[0] > 4', )),
             count=True, store_file=True,
             file_name='tbl_given-filename.txt'),
        dict(condition=dict(All=('ev: ev.njets[0] > 4', )),
             count=True, store_file=True,
             file_name='tbl_given-filename.txt',
             file_path='tbl_given-filename.txt')
    ),

    #
    (
        dict(condition=dict(All=('ev: ev.njets[0] > 4', )),
             count=True, store_file=True,
             file_dir='out',
             file_name='tbl_given-filename.txt'),
        dict(condition=dict(All=('ev: ev.njets[0] > 4', )),
             count=True, store_file=True,
             file_dir='out',
             file_name='tbl_given-filename.txt',
             file_path='out/tbl_given-filename.txt')
    ),

    #
    (
        dict(condition=dict(All=('ev: ev.njets[0] > 4', )),
             count=True, store_file=True,
             file_dir='out',
             file_name='tbl_given-filename-1.txt',
             file_path='abc/def/tbl_given-filename-2.txt'),
        dict(condition=dict(All=('ev: ev.njets[0] > 4', )),
             count=True, store_file=True,
             file_dir='out',
             file_name='tbl_given-filename-1.txt',
             file_path='abc/def/tbl_given-filename-2.txt'),
    ),

]

@pytest.mark.parametrize('arg, expected', params)
def test_complete(arg, expected):
    actual = complete_selection_cfg(arg)
    assert expected == actual
    assert arg is not actual

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import os

from .._misc import is_dict

##__________________________________________________________________||
def complete_selection_cfg(cfg):
    """complete a selection config.

    Parameters
    ----------
    cfg : dict
        A table config

    Returns
    -------
    dict
        A completed selection config

    """

    default_cfg = dict(
        count=False,
        store_file=False
    )

    default_file_name = 'tbl_selection_count.txt'

    cfg = _wrap_cfg(cfg)

    ret = default_cfg
    ret.update(cfg)

    if ret['store_file']:
        if 'file_path' not in ret:
            file_dir = ret.get('file_dir', '')
            file_name = ret.get('file_name', default_file_name)
            ret['file_path'] = os.path.join(file_dir, file_name)

    return ret

##__________________________________________________________________||
def _wrap_cfg(cfg):

    if isinstance(cfg, str):
        return dict(condition=cfg)

    if is_dict(cfg):
        if 'condition' not in cfg:
            return dict(condition=cfg)

    return cfg
##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import os
import numpy as np
import pandas as pd

import pytest

from pandas.util.testing import assert_frame_equal

import alphatwirl

from qtwirl import qtwirl

##__________________________________________________________________||
pytestmark = pytest.mark.filterwarnings('ignore::RuntimeWarning')

##__________________________________________________________________||
TESTDATADIR = os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))), 'data')

SAMPLE_ROOT_FILE_DIR = os.path.join(TESTDATADIR, 'root')
SAMPLE_ROOT_FILE_NAMES = [
    'sample_chain_01.root',
    'sample_chain_02.root',
    'sample_chain_03_zombie.root',
    'sample_chain_04.root',
]
SAMPLE_ROOT_FILE_PATHS = [os.path.join(SAMPLE_ROOT_FILE_DIR, n) for n in SAMPLE_ROOT_FILE_NAMES]

##__________________________________________________________________||
@pytest.fixture()
def file_dir(tmpdir_factory):
   ret = str(tmpdir_factory.mktemp(''))
   return ret

@pytest.mark.parametrize('store_file', [False, True])
def test_one_table(store_file, file_dir):

    ##
    sample_paths = SAMPLE_ROOT_FILE_PATHS

    ##
    RoundLog = alphatwirl.binning.RoundLog
    reader_cfg = dict(
        key_name='jet_pt',
        key_binning=RoundLog(0.1, 100),
        key_index='*',
        key_out_name='jet_pt',
        store_file=store_file,
        file_dir=file_dir
    )

    results = qtwirl(
        sample_paths,
        reader_cfg=reader_cfg,
        tree_name='tree',
        process=16,
        quiet=False,
        max_events_per_process=500
    )

    ##
    tbl_expected_dir = os.path.join(TESTDATADIR, 'tbl')
    tbl_expected_path = os.path.join(tbl_expected_dir, '00', 'tbl_n.jet_pt-w.txt')
    tbl_expected = pd.read_table(tbl_expected_path, delim_whitespace=True)

    ##
    assert_frame_equal(tbl_expected, results, check_names=True)

    ##
    if store_file is True:
        tbl_stored_file_name = 'tbl_n.jet_pt-w.txt'
        assert [tbl_stored_file_name] == os.listdir(file_dir)
        tbl_stored_path = os.path.join(file_dir, tbl_stored_file_name)
        tbl_stored = pd.read_table(tbl_stored_path, delim_whitespace=True)
        assert_frame_equal(tbl_expected, tbl_stored, check_names=True)
    else:
        assert [ ] == os.listdir(file_dir)

##__________________________________________________________________||
@pytest.mark.parametrize('count', [False, True])
@pytest.mark.parametrize('store_file', [False, True])
def test_one_selection_four_tables(count, store_file, file_dir):

    ##
    sample_paths = SAMPLE_ROOT_FILE_PATHS

    ##
    RoundLog = alphatwirl.binning.RoundLog
    from scribblers.essentials import FuncOnNumpyArrays
    reader_cfg = [
        dict(selection_cfg=dict(
            condition=dict(All=('ev: ev.njets[0] > 4', )),
            count=count,
            store_file=store_file,
            file_dir=file_dir
        )),
        dict(key_name='jet_pt',
             key_binning=RoundLog(0.1, 100),
             key_index='*',
             key_out_name='jet_pt',
             store_file=store_file,
             file_dir=file_dir
        ),
        dict(key_name='met',
             key_binning=RoundLog(0.1, 100),
             store_file=store_file,
             file_dir=file_dir
        ),
        dict(
            key_name=('njets', 'met'),
            key_binning=(None, RoundLog(0.2, 100, min=50, underflow_bin=0)),
             store_file=store_file,
             file_dir=file_dir
        ),
        dict(reader=FuncOnNumpyArrays(
            src_arrays=['jet_pt'],
            out_name='ht',
            func=np.sum)),
        dict(key_name='ht',
             key_binning=RoundLog(0.1, 100),
             store_file=store_file,
             file_dir=file_dir
        ),
    ]

    results = qtwirl(
        sample_paths,
        reader_cfg=reader_cfg,
        tree_name='tree',
        process=16,
        quiet=False,
        max_events_per_process=500
    )

    ##
    tbl_expected_file_names = [
        'tbl_selection_count.txt',
        'tbl_n.jet_pt-w.txt',
        'tbl_n.met.txt',
        'tbl_n.njets.met.txt',
        'tbl_n.ht.txt',
    ]
    if count is False:
        tbl_expected_file_names = tbl_expected_file_names[1:]

    ##
    assert len(tbl_expected_file_names) == len(results)

    ##
    tbl_expected_dir = os.path.join(TESTDATADIR, 'tbl')
    tbl_expected_paths = [os.path.join(tbl_expected_dir, '01', n) for n in tbl_expected_file_names]
    tbls_expected = [pd.read_table(p, delim_whitespace=True) for p in tbl_expected_paths]

    ##
    for expected, actual in zip(tbls_expected, results):
        assert_frame_equal(expected, actual, check_names=True, check_less_precise=True)


    ##
    if store_file is True:
        assert set(tbl_expected_file_names) == set(os.listdir(file_dir))
        for expected, file_name in zip(tbls_expected, tbl_expected_file_names):
            tbl_stored_path = os.path.join(file_dir, file_name)
            tbl_stored = pd.read_table(tbl_stored_path, delim_whitespace=True)
            assert_frame_equal(expected, tbl_stored, check_names=True, check_less_precise=True)

    else:
        assert [ ] == os.listdir(file_dir)

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import pytest

try:
    import unittest.mock as mock
except ImportError:
    import mock

from qtwirl._parser.tableconfig import complete_table_cfg

from alphatwirl.summary import Count, WeightCalculatorOne


def eq(self, other):
    return isinstance(other, self.__class__)
WeightCalculatorOne.__eq__ = eq

##__________________________________________________________________||
class MockSummary2:
    pass

class MockBinning:
    pass

defaultWeight = WeightCalculatorOne()
binning1 = MockBinning()
binning2 = MockBinning()

##__________________________________________________________________||
KEYS_NEED_TO_EXIST_BUT_DONT_TEST_VALUES = (
    'key_name', 'key_index', 'key_binning', 'key_out_name',
    'val_name', 'val_index', 'agg_class', 'agg_name',
    'weight', 'sort', 'nevents')

params = [

    pytest.param(
        dict(),
        dict(store_file=False),
        id='empty'),

    pytest.param(
        dict(store_file=False),
        dict(store_file=False),
        id='empty-false'),

    pytest.param(
        dict(store_file=True),
        dict(
            store_file=True,
            file_path='tbl_n.txt',
        ),
        id='empty-true'),

    pytest.param(
        dict(key_name='met', store_file=True),
        dict(
            store_file=True,
            file_path='tbl_n.met.txt',
        ),
        id='composed'),

    pytest.param(
        dict(key_name='met', store_file=True, file_dir='out'),
        dict(
            store_file=True,
            file_dir='out',
            file_path='out/tbl_n.met.txt',
        ),
        id='composed-file-dir'),

    pytest.param(
        dict(
            key_name='met',
            store_file=True,
            file_name='tbl.given-filename.txt',
        ),
        dict(
            store_file=True,
            file_name='tbl.given-filename.txt',
            file_path='tbl.given-filename.txt',
        ),
        id='given'),

    pytest.param(
        dict(
            key_name='met',
            store_file=True,
            file_dir='out',
            file_name='tbl.given-filename.txt',
        ),
        dict(
            store_file=True,
            file_dir='out',
            file_name='tbl.given-filename.txt',
            file_path='out/tbl.given-filename.txt',
        ),
        id='given'),

    pytest.param(
        dict(
            key_name='met',
            store_file=True,
            file_dir='out',
            file_name='tbl.given-filename_1.txt',
            file_path='abc/def/tbl.given-filename_2.txt',
        ),
        dict(
            store_file=True,
            file_dir='out',
            file_name='tbl.given-filename_1.txt',
            file_path='abc/def/tbl.given-filename_2.txt',
        ),
        id='given'),

    pytest.param(
        dict(
            key_name='met',
            store_file=True,
            file_name_prefix='tbl_abc',
        ),
        dict(
            store_file=True,
            file_name_prefix='tbl_abc',
            file_path='tbl_abc.met.txt',
        ),
        id='prefix'),

]

@pytest.mark.parametrize('arg, expected', params)
def test_complete(arg, expected):
    actual = complete_table_cfg(arg)
    assert arg is not actual
    for k in KEYS_NEED_TO_EXIST_BUT_DONT_TEST_VALUES:
        actual.pop(k)
    assert expected == actual

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import functools

import alphatwirl

from ._parser.inputs import parse_data
from ._parser.readerconfig import expand_reader_config
from ._builder.func import build_reader, create_file_loaders, let_reader_read

##__________________________________________________________________||
__all__ = ['qtwirl']

##__________________________________________________________________||
def qtwirl(data, reader_cfg,
           tree_name=None,
           parallel_mode='multiprocessing',
           dispatcher_options=None,
           process=4, quiet=False,
           user_modules=None,
           max_events=-1, max_files=-1,
           max_events_per_process=-1, max_files_per_process=1,
           skip_error_files=True):
    """qtwirl (quick-twirl), one-function interface to alphatwirl

    Summarize event data in ``file`` in the way specified by
    ``reader_cfg`` and return the results.

    Parameters
    ----------
    data : str or list of str
        Input file path(s)
    reader_cfg : dict or list of dict
        Reader configuration
    parallel_mode : str, optional
        "multiprocessing" (default) or "htcondor"
    dispatcher_options : dict, optional
        Options to dispatcher
    process : int, optional
        The number of processes when ``parallel_mode`` is
        "multiprocessing"
    quiet : bool, optional
    user_modules : list, optional
        The names of modules to be sent to worker nodes when
        parallel_mode is "htcondor"
    max_events : int, optional
    max_files : int, optional
    max_events_per_process : int, optional
    max_files_per_process : int, optional
    skip_error_files, bool, default True

    Returns
    -------
    DataFrame or list of DataFrame
        Summary of event data

    """

    ##
    files = parse_data(data)

    ##
    reader_cfg = expand_reader_config(reader_cfg)
    reader = build_reader(reader_cfg)

    ##
    if dispatcher_options is None:
        dispatcher_options=dict()


    ##
    default_user_modules = ('qtwirl', 'alphatwirl')
    if user_modules is None:
        user_modules = ()
    user_modules = set(user_modules)
    user_modules.update(default_user_modules)

    ##
    parallel = alphatwirl.parallel.build_parallel(
        parallel_mode=parallel_mode, quiet=quiet,
        processes=process,
        user_modules=user_modules,
        dispatcher_options=dispatcher_options)
    func_create_file_loaders = functools.partial(
        create_file_loaders,
        tree_name=tree_name,
        max_events=max_events, max_events_per_run=max_events_per_process,
        max_files=max_files, max_files_per_run=max_files_per_process,
        check_files=True, skip_error_files=skip_error_files)
    read_files = functools.partial(
        let_reader_read, reader=reader, parallel=parallel,
        func_create_file_loaders=func_create_file_loaders)

    parallel.begin()
    ret = read_files(files=files)
    parallel.end()

    if isinstance(reader, alphatwirl.loop.ReaderComposite):
        ret = [r for r in ret if r is not None]
    return ret

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
"""

A generic tool to expand config written in dict and a list of dict.

This module is not specific to qtwirl and can become an independent
package in the future.


"""

import logging
import functools
import pprint

from .._misc import is_dict

##__________________________________________________________________||
__all__ = ['config_expander']

##__________________________________________________________________||
def config_expander(expand_func_map=None, config_keys=None,
                    default_config_key=None, default_cfg_dict=None):
    """return a function that expands a config

    Parameters
    ----------
    expand_func_map : dict, optional
        A map from a config key to a function 1) that expands the
        config for the key or 2) that updates the shared objects

    config_keys : list, optional
        A list of extra config keys that are not in keys of
        ``expand_func_map``.

    default_config_key: str, optional
        A default key

    default_cfg_dict: dict, optional
        A dict of default config

    Returns
    -------
    function
        A function that expands a config

    """

    #
    if expand_func_map is None:
        expand_func_map = {}

    if config_keys is None:
        config_keys = []

    if default_cfg_dict is None:
        default_cfg_dict = {}

    #
    expand_func_map = expand_func_map.copy() # so as not to modify
                                                 # the original

    #
    if 'set_default' in expand_func_map:
        logger = logging.getLogger(__name__)
        msg = '"set_default" is in expand_func_map: expand_func_map={!r}'.format(expand_func_map)
        logger.info(msg)
    else:
        expand_func_map['set_default'] = _set_default

    #
    config_keys = set(config_keys)
    config_keys.update(expand_func_map.keys())

    if default_config_key is not None:
        config_keys.add(default_config_key)

    #
    default_cfg_dict = _wrap_default_cfg(default_cfg_dict, config_keys, default_config_key)
    default_cfg_stack = [default_cfg_dict]

    #
    shared = dict(
        default_cfg_stack=default_cfg_stack,
        expand_func_map=expand_func_map,
        config_keys=config_keys,
        default_config_key=default_config_key,
    )

    #
    func_apply_default = functools.partial(
        _apply_default_for_one_key, shared=shared)
    functools.update_wrapper(func_apply_default, _apply_default_for_one_key)
    func_apply_default.__name__ = 'apply_default'
    shared['func_apply_default'] = func_apply_default

    #
    ret = functools.partial(_expand_config, shared=shared)
    functools.update_wrapper(ret, _expand_config)
    ret.__name__ = 'expand_config'

    return ret

##__________________________________________________________________||
def _expand_config(cfg, shared=None):
    """expand a config into its full form

    Parameters
    ----------
    cfg : dict, None, or list of dicts and None
        Configuration

    shared : dict, optional

        To be given by ``config_expander()`` with
        ``functools.partial()``.

    Returns
    -------
    dict, list of dicts, or None
        Configuration in its full form

    """

    if cfg is None:
        return None

    if shared is None:
        shared = {}

    if is_dict(cfg):
        cfg = _expand_one_dict(cfg, shared)
        if not cfg:
            return None
        return cfg

    # cfg is a list of dicts and None

    ret = [ ]
    for c in cfg:
        if c is None:
            continue

        c = _expand_one_dict(c, shared)

        if c is None:
            continue

        if isinstance(c, list):
            c = [e for e in c if e is not None]
            ret.extend(c)
        else:
            ret.append(c)

    return ret

##__________________________________________________________________||
def _expand_one_dict(cfg, shared):
    """expand a piece of config

    Parameters
    ----------
    cfg : dict
        Configuration

    shared : dict
        A dict of shared objects

    Returns
    -------
    dict, list
        Expanded configuration

    """

    if shared['default_config_key'] is not None:
        if not (len(cfg) == 1 and list(cfg.keys())[0] in shared['config_keys']):
            cfg = {shared['default_config_key']: cfg}

    if not len(cfg) == 1:
        return cfg.copy()

    key, val = list(cfg.items())[0]

    if key not in shared['config_keys']:
        cfg = _apply_default_for_all_keys(cfg, shared)
        return cfg.copy()

    if key not in shared['expand_func_map']:
        cfg = _apply_default_for_all_keys(cfg, shared)
        return cfg.copy()

    expand_func = shared['expand_func_map'][key]
    try:
        return expand_func(val, shared)
    except TypeError:
        return expand_func(val)


##__________________________________________________________________||
def _set_default(cfg, shared):
    wrapped = _wrap_default_cfg(
        cfg, shared['config_keys'], shared['default_config_key'])
    shared['default_cfg_stack'].append(wrapped)
    return None

def _wrap_default_cfg(cfg, config_keys, default_config_key):
    if default_config_key is None:
        return cfg
    if set(cfg.keys()) <= set(config_keys):
        return cfg
    return {default_config_key: cfg}

##__________________________________________________________________||
def _apply_default_for_all_keys(cfg, shared):
    if 'default_cfg_stack' not in shared:
        return cfg
    ignore = ('set_default', )
    ret = {}
    for key, val in cfg.items():
        if key in ignore:
            ret[key] = val
            continue
        ret[key] = _apply_default_for_one_key(key, val, shared)
    return ret

def _apply_default_for_one_key(key, cfg, shared):
    """apply default to a config for a key

    Parameters
    ----------
    key : str
        A config key
    cfg : dict
        A config
    shared : dict
        A dict of shared objects.

    Returns
    -------
    function
        A config with default applied

    """

    ret = {}
    for default_cfg in shared['default_cfg_stack']:
        ret.update(default_cfg.get(key, {}))
    ret.update(cfg)
    return ret

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import os
import copy
import collections
import functools

import pandas as pd

import alphatwirl
from alphatwirl.roottree.inspect import get_entries_in_tree_in_file
from alphatwirl.loop.splitfuncs import create_files_start_length_list
from alphatwirl.loop.merge import merge_in_order

from .._misc import is_dict

##__________________________________________________________________||
def build_reader(cfg):
    """build a reader based on a configuration

    Parameters
    ----------
    cfg : dict or list of dicts
        Reader configuration (in its full form)

    Returns
    -------
    object
        A reader, which, for example, can be given to ``EventLoop`` of
        AlphaTwirl

    """
    if is_dict(cfg):
        return _build_single_reader(cfg)

    # cfg is a list
    readers = [_build_single_reader(c) for c in cfg]
    ret = alphatwirl.loop.ReaderComposite(readers=readers)
    return ret

def _build_single_reader(cfg):
    # cfg is a dict with one item
    key, val = list(cfg.items())[0]
    if key == 'table_cfg':
        return build_reader_for_table_config(val)
    elif key == 'selection_cfg':
        return build_reader_for_selection_config(val)
    elif key == 'reader':
        return val
    else:
        # TODO: produce warnings here
        return None

##__________________________________________________________________||
def build_reader_for_selection_config(cfg):
    if cfg['count'] is True:
        ret = alphatwirl.selection.build_selection(
            path_cfg=cfg['condition'],
            AllClass=alphatwirl.selection.modules.AllwCount,
            AnyClass=alphatwirl.selection.modules.AnywCount,
            NotClass=alphatwirl.selection.modules.NotwCount
        )
        ret.collector = build_collector_for_selection_config(cfg)
    else:
        ret = alphatwirl.selection.build_selection(path_cfg=cfg['condition'])
    return ret

def build_collector_for_selection_config(cfg):
    columns = ('depth', 'class', 'name', 'pass', 'total')
    if cfg['store_file'] is True:
        return functools.partial(
            to_dataframe_store_file,
            columns=columns, path=cfg['file_path'])
    else:
        return functools.partial(to_dataframe, columns=columns)

##__________________________________________________________________||
def build_reader_for_table_config(cfg):

    ## binnings: replace None with Echo with no next func
    ## TODO: this feature should be included in KeyValueComposer
    ## https://github.com/alphatwirl/alphatwirl/issues/49
    echo = alphatwirl.binning.Echo(nextFunc=None)
    binnings = cfg['key_binning']
    if binnings:
        binnings = tuple(b if b else echo for b in binnings)

    ##
    keyValComposer = alphatwirl.summary.KeyValueComposer(
        keyAttrNames=cfg['key_name'],
        binnings=binnings,
        keyIndices=cfg['key_index'],
        valAttrNames=cfg['val_name'],
        valIndices=cfg['val_index']
    )

    ##
    nextKeyComposer = alphatwirl.summary.NextKeyComposer(binnings) if binnings is not None else None

    ##
    summarizer = alphatwirl.summary.Summarizer(Summary=cfg['agg_class'])

    ##
    collector = build_collector_for_table_config(cfg)

    ##
    reader = alphatwirl.summary.Reader(
        keyValComposer=keyValComposer,
        summarizer=summarizer,
        collector=collector,
        nextKeyComposer=nextKeyComposer,
        weightCalculator=cfg['weight'],
        nevents=cfg['nevents']
    )
    return reader

##__________________________________________________________________||
def build_collector_for_table_config(cfg):
    columns = cfg['key_out_name'] + cfg['agg_name']
    if cfg['store_file'] is True:
        return functools.partial(
            to_dataframe_store_file,
            columns=columns, path=cfg['file_path'])
    else:
        return functools.partial(to_dataframe, columns=columns)

##__________________________________________________________________||
def to_dataframe_store_file(reader, columns, path):
    tuple_list = reader.results().to_tuple_list()

    if tuple_list is None:
        tuple_list = [ ]

    tuple_list_with_header = [columns] + tuple_list

    dir_ = os.path.dirname(path)
    alphatwirl.misc.mkdir_p(dir_)
    with open(path, 'w') as f:
        content = alphatwirl.misc.list_to_aligned_text(tuple_list_with_header)
        f.write(content)

    return pd.DataFrame(tuple_list, columns=columns)

def to_dataframe(reader, columns):
    tuple_list = reader.results().to_tuple_list()
    if tuple_list is None:
        tuple_list = [ ]
    return pd.DataFrame(tuple_list, columns=columns)

##__________________________________________________________________||
def create_file_loaders(
        files, tree_name,
        max_events=-1, max_events_per_run=-1,
        max_files=-1, max_files_per_run=1,
        check_files=True, skip_error_files=False):

        func_get_nevents_in_file = functools.partial(
            get_entries_in_tree_in_file,
            tree_name=tree_name,
            raises=not skip_error_files
        )

        files_start_length_list = create_files_start_length_list(
            files,
            func_get_nevents_in_file=func_get_nevents_in_file,
            max_events=max_events,
            max_events_per_run=max_events_per_run,
            max_files=max_files,
            max_files_per_run=max_files_per_run
        )
        # list of (files, start, length), e.g.,
        # [
        #     (['A.root'], 0, 80),
        #     (['A.root', 'B.root'], 80, 80),
        #     (['B.root'], 60, 80),
        #     (['B.root', 'C.root'], 140, 80),
        #     (['C.root'], 20, 10)
        # ]

        ret = [ ]
        for files, start, length in files_start_length_list:
            config = dict(
                events_class=alphatwirl.roottree.BEvents,
                file_paths=files,
                tree_name=tree_name,
                max_events=length,
                start=start,
                check_files=check_files,
                skip_error_files=skip_error_files,
            )
            ret.append(alphatwirl.roottree.BuildEvents(config))
        return ret

##__________________________________________________________________||
def let_reader_read(files, reader, parallel, func_create_file_loaders):
    eventLoopRunner = alphatwirl.loop.MPEventLoopRunner(parallel.communicationChannel)
    eventLoopRunner.begin()

    file_loaders = func_create_file_loaders(files)
    njobs = len(file_loaders)
    eventLoops = [ ]
    for i, file_loader in enumerate(file_loaders):
        reader_copy = copy.deepcopy(reader)
        eventLoop = alphatwirl.loop.EventLoop(file_loader, reader_copy, '{} / {}'.format(i, njobs))
        eventLoops.append(eventLoop)
    runids = eventLoopRunner.run_multiple(eventLoops)
    # e.g., [0, 1, 2]

    runid_reader_map = collections.OrderedDict([(i, None) for i in runids])
    # e.g., OrderedDict([(0, None), (1, None), (2, None)])

    runids_towait = runids[:]
    while runids_towait:
        runid, reader_returned = eventLoopRunner.receive_one()
        merge_in_order(runid_reader_map, runid, reader_returned)
        runids_towait.remove(runid)

    if runid_reader_map:
        # assert 1 == len(runid_reader_map)
        reader = list(runid_reader_map.values())[0]
    return reader.collect()

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import sys
import functools
import pytest

try:
    import unittest.mock as mock
except ImportError:
    import mock

from qtwirl._parser.expander import config_expander

##__________________________________________________________________||
params = [
    pytest.param(
        dict(),
        set([]),
        id='empty'),

    pytest.param(
        dict(expand_func_map={'abc_cfg': mock.sentinel.expand_abc_cfg}),
        set(['abc_cfg']),
        id='empty-config-keys'),

    pytest.param(
        dict(
            expand_func_map={'abc_cfg': mock.sentinel.expand_abc_cfg},
            config_keys=['def_cfg']
        ),
        set(['abc_cfg', 'def_cfg']),
        id='default-none'),

    pytest.param(
        dict(
            expand_func_map={'abc_cfg': mock.sentinel.expand_abc_cfg},
            config_keys=['def_cfg'], default_config_key='abc_cfg'
        ),
        set(['abc_cfg', 'def_cfg']),
        id='default-in-list'),

    pytest.param(
        dict(
            expand_func_map={'abc_cfg': mock.sentinel.expand_abc_cfg},
            config_keys=['def_cfg'], default_config_key='xyz_cfg'
        ),
        set(['abc_cfg', 'def_cfg', 'xyz_cfg']),
        id='default-not-in-list'),
]

@pytest.mark.parametrize('kwargs, expected', params)
def test_config_expander_config_keys(
        kwargs, expected) :
    """test if config_keys are correctly initialized

    """
    expected.add('set_default')
    expand_config = config_expander(**kwargs)
    actual_shared = expand_config.keywords['shared']
    actual = actual_shared['config_keys']
    assert expected == actual

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import os
import functools
import alphatwirl

##__________________________________________________________________||
def complete_table_cfg(cfg):
    """complete a table config.

    Parameters
    ----------
    cfg : dict
        A table config

    Returns
    -------
    dict
        A completed table config

    """

    default_agg_class = alphatwirl.summary.Count
    default_agg_name_for_default_agg_class = ('n', 'nvar')

    default_weight = alphatwirl.summary.WeightCalculatorOne()

    default_cfg = dict(
        key_name=( ),
        key_index=None,
        key_binning=None,
        val_name=None,
        val_index=None,
        agg_class=default_agg_class,
        weight=default_weight,
        sort=True,
        nevents=None,
        store_file=False,
    )

    func_compose_tbl_filename = functools.partial(
        compose_tbl_filename_from_config,
        default=dict(file_name_prefix='tbl_n'))

    ret = default_cfg
    ret.update(cfg)

    ret['key_out_name'] = ret.get('key_out_name', ret['key_name'])

    if isinstance(ret['key_name'], str):
        ret['key_name'] = (ret['key_name'], )
        ret['key_out_name'] = (ret['key_out_name'], )
        if ret['key_index'] is not None:
            ret['key_index'] = (ret['key_index'], )
        if ret['key_binning'] is not None:
            ret['key_binning'] = (ret['key_binning'], )

    if isinstance(ret['val_name'], str):
        ret['val_name'] = (ret['val_name'], )
        if ret['val_index'] is not None:
            ret['val_index'] = (ret['val_index'], )

    use_default_agg_class = 'agg_class' not in cfg
    if 'agg_name' not in ret:
        if use_default_agg_class:
            ret['agg_name'] = default_agg_name_for_default_agg_class
        else:
            ret['agg_name'] = ret['val_name'] if ret['val_name'] is not None else ()

    if isinstance(ret['agg_name'], str):
        ret['agg_name'] = (ret['agg_name'], )

    if ret['store_file']:
        if 'file_path' not in ret:
            file_dir = ret.get('file_dir', '')
            file_name = ret.get('file_name', func_compose_tbl_filename(ret))
            ret['file_path'] = os.path.join(file_dir, file_name)

    return ret

##__________________________________________________________________||
def compose_tbl_filename_from_config(table_cfg, default=None):

    """compose a file name based on a table config.

    This function calls ``compose_tbl_filename()``.

    Parameters
    ----------
    table_cfg : dict
        A table config
    default : dict, optional
        A default config

    Returns
    -------
    str
        A file name

    """

    default_base = dict(
        file_name_prefix='tbl',
        file_name_suffix='txt',
        file_name_var_separator='.',
        file_name_idx_separator='-',
    )

    if default is None:
        default = default_base
    else:
        default_base.update(default)
        default = default_base

    cfg = default.copy()
    cfg.update(table_cfg)

    ret = compose_tbl_filename(
        cfg['key_out_name'],
        key_indices=cfg.get('key_index', None),
        prefix=cfg['file_name_prefix'],
        suffix=cfg['file_name_suffix'],
        var_separator=cfg['file_name_var_separator'],
        idx_separator=cfg['file_name_idx_separator']
    )

    return ret

##__________________________________________________________________||
def compose_tbl_filename(
        key_names, key_indices=None,
        prefix='tbl', suffix='txt',
        var_separator='.', idx_separator='-'):
    """compose a file name based on key names

    Parameters
    ----------
    key_names : list of str
        A list of key names
    key_indices : list of str
        A list of key indices, which can be a number, ``None``, a
        wildcard ``*``, a wildcards in parentheses ``(*)``, a back
        reference ``\\n`` (``n`` is a number),
        e.g., ``(1, None, '*', '(*)', '\\1')``
    prefix : str, default ``tbl``
        A prefix of a file name
    suffix : str, default ``txt``
        A suffix of a file name (filename extension)
    var_separator : str, default '.'
        A separator between key names
    idx_separator : str, default '-'
        A separator between a key name and a key index

    Returns
    -------
    str
        A file name

    """

    if not key_names:
        return prefix + '.' + suffix # e.g. "tbl.txt"

    if key_indices is None:
        colidxs = key_names
        # e.g., ('var1', 'var2', 'var3'),

        middle = var_separator.join(colidxs)
        # e.g., 'var1.var2.var3'

        ret = prefix + var_separator + middle + '.' + suffix
        # e.g., 'tbl.var1.var2.var3.txt'

        return ret

    # e.g.,
    # key_names = ('var1', 'var2', 'var3', 'var4', 'var5'),
    # key_indices = (1, None, '*', '(*)', '\\1')

    idx_str = key_indices
    # e.g., (1, None, '*', '(*)', '\\1')

    idx_str = ['w' if i == '*' else i for i in idx_str]
    # e..g, [1, None, 'w', '(*)', '\\1']

    idx_str = ['wp' if i == '(*)' else i for i in idx_str]
    # e.g., [1, None, 'w', 'wp', '\\1']

    idx_str = ['b{}'.format(i[1:]) if isinstance(i, str) and i.startswith('\\') else i for i in idx_str]
    # e.g., [1, None, 'w', 'wp', 'b1']

    idx_str = ['' if i is None else '{}{}'.format(idx_separator, i) for i in idx_str]
    # e.g., ['-1', '', '-w', '-wp', '-b1']

    colidxs = [n + i for n, i in zip(key_names, idx_str)]
    # e.g., ['var1-1', 'var2', 'var3-w', 'var4-wp', 'var5-b1']

    middle = var_separator.join(colidxs)
    # e.g., 'var1-1.var2.var3-w.var4-wp.var5-b1'

    ret =  prefix + var_separator + middle + '.' + suffix
    # e.g., tbl.var1-1.var2.var3-w.var4-wp.var5-b1.txt

    return ret

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import functools

from .expander import config_expander

from .tableconfig import complete_table_cfg
from .selectionconfig import complete_selection_cfg

##__________________________________________________________________||
def expand_table_cfg(cfg):
    return dict(table_cfg=complete_table_cfg(cfg))

def expand_selection_cfg(cfg):
    return dict(selection_cfg=complete_selection_cfg(cfg))

def expand_reader(reader):
    if isinstance(reader, list) or isinstance(reader, tuple):
        return [dict(reader=r) for r in reader if r is not None]
    return dict(reader=reader)

expand_func_map = {
    'table_cfg': expand_table_cfg,
    'selection_cfg': expand_selection_cfg,
    'reader': expand_reader,
}

default_config_key = 'table_cfg'

##__________________________________________________________________||
expand_reader_config = config_expander(
    expand_func_map=expand_func_map,
    default_config_key=default_config_key
)

##__________________________________________________________________||
